LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

ENTITY pwm_fan_controller IS
   GENERIC(
      p_clkin_freq            : positive := 100_000_000; -- Frequency of i_clk in Hz
      p_pwm_freq              : positive := 30; -- Frequency of PWM signal in Hz
      p_pwm_precision         : positive := 255; -- Stepsize of PWM signal
      p_pulse_stretch_timeout : positive := 100_000_000 -- Timeout for pulse_stretching (if not tacho pulses are received)
   );
   PORT(
      i_clk           : IN  std_logic;
      i_rst           : IN  std_logic;
      i_fan_enable    : IN  std_logic;
      i_fan_req_speed : IN  std_logic_vector(7 DOWNTO 0); -- requested speed (in % duty cycle)
      o_fan_act_speed : OUT std_logic_vector(31 DOWNTO 0); -- actual speed (in tacho pulses)
      o_fan_pwm       : OUT std_logic;
      i_fan_tacho     : IN  std_logic;
      o_fan_error     : OUT std_logic
   );
END ENTITY;

ARCHITECTURE rtl OF pwm_fan_controller IS

   CONSTANT c_pwm_div_cnt : positive := p_clkin_freq / (p_pwm_freq * p_pwm_precision) / 2;

   SIGNAL r_pwm_clk        : std_logic := '0';
   SIGNAL r_pwm_duty_cycle : natural RANGE 0 TO p_pwm_precision;

   SIGNAL r_tacho_sync        : std_logic_vector(7 DOWNTO 0) := (OTHERS => '1');
   SIGNAL n_tacho_rising_edge : std_logic;
   SIGNAL r_do_speed_check    : std_logic                    := '0';

   SIGNAL r_fan_pwm : std_logic;

   SIGNAL r_cycle_cnt : natural RANGE 0 TO p_pwm_freq - 1 := 0;

   TYPE t_tacho_cnt_states IS (IDLE, WAIT_FOR_FIRST_TACHO_EDGE, WAIT_FOR_SECOND_TACHO_EDGE, CALC_AVG);
   SIGNAL sm_tacho_cnt : t_tacho_cnt_states := IDLE;

BEGIN

   PROCESS(i_clk)
      VARIABLE v_div_cnt : natural RANGE 0 TO c_pwm_div_cnt - 1;
   BEGIN
      IF (rising_edge(i_clk)) THEN
         IF (i_rst = '1') THEN
            v_div_cnt := c_pwm_div_cnt - 1;
         ELSE
            -- Create lower clock for creating PWM signal
            IF (v_div_cnt = 0) THEN
               r_pwm_clk <= NOT r_pwm_clk;
               v_div_cnt := c_pwm_div_cnt - 1;
            ELSE
               v_div_cnt := v_div_cnt - 1;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   r_pwm_duty_cycle <= to_integer(unsigned(i_fan_req_speed));
   o_fan_pwm        <= '1' WHEN (sm_tacho_cnt /= IDLE OR i_rst = '1') ELSE r_fan_pwm;

   PROCESS(r_pwm_clk)
      VARIABLE v_pwm_period_cnt : natural RANGE 0 TO p_pwm_precision := 0;
   BEGIN
      IF (rising_edge(r_pwm_clk)) THEN
         IF (i_rst = '1') THEN
            r_fan_pwm   <= '0';
            r_cycle_cnt <= p_pwm_freq - 1;
         ELSE

            IF (i_fan_enable = '1') THEN

               -- Count from 0 to p_pwm_precision continuously 
               IF (v_pwm_period_cnt >= p_pwm_precision) THEN
                  v_pwm_period_cnt := 0;

                  IF (r_cycle_cnt > 0) THEN
                     r_cycle_cnt      <= r_cycle_cnt - 1;
                     r_do_speed_check <= '0';
                  ELSE
                     r_cycle_cnt      <= p_pwm_freq - 1;
                     r_do_speed_check <= '1';
                  END IF;

               ELSE
                  v_pwm_period_cnt := v_pwm_period_cnt + 1;
               END IF;

               -- Create PWM signal
               IF (v_pwm_period_cnt <= r_pwm_duty_cycle) THEN
                  r_fan_pwm <= '1';
               ELSE
                  r_fan_pwm <= '0';
               END IF;
            ELSE
               r_fan_pwm <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;

   n_tacho_rising_edge <= '1' WHEN r_tacho_sync = (r_tacho_sync'high => '0', r_tacho_sync'high - 1 DOWNTO 0 => '1') ELSE '0';

   PROCESS(i_clk)
      VARIABLE v_tacho_period_cnt    : unsigned(31 DOWNTO 0) := (OTHERS => '0');
      VARIABLE v_do_speed_check_prev : std_logic             := '0';

      CONSTANT c_avg_length : natural := 8;
      TYPE t_unsigned32_array IS ARRAY (0 TO c_avg_length - 1) OF unsigned(31 DOWNTO 0);

      VARIABLE v_measurements      : t_unsigned32_array                  := (OTHERS => (OTHERS => '0'));
      VARIABLE v_measurement_index : natural RANGE 0 TO c_avg_length - 1 := 0;
      VARIABLE v_sum               : unsigned(31 DOWNTO 0)               := (OTHERS => '0');

   BEGIN
      IF (rising_edge(i_clk)) THEN
         IF (i_rst = '1') THEN
            o_fan_error           <= '0';
            sm_tacho_cnt          <= IDLE;
            o_fan_act_speed       <= (OTHERS => '0');
            r_tacho_sync          <= (OTHERS => '1');
            v_do_speed_check_prev := '0';
            v_measurement_index   := 0;
         ELSE

            r_tacho_sync <= r_tacho_sync(6 DOWNTO 0) & i_fan_tacho;

            IF (i_fan_enable = '1') THEN

               CASE (sm_tacho_cnt) IS
                  WHEN IDLE =>
                     IF (r_do_speed_check = '1' AND v_do_speed_check_prev = '0') THEN
                        sm_tacho_cnt       <= WAIT_FOR_FIRST_TACHO_EDGE;
                        v_tacho_period_cnt := (OTHERS => '0');
                     END IF;
                  -----------------------------------
                  WHEN WAIT_FOR_FIRST_TACHO_EDGE =>

                     IF (v_tacho_period_cnt >= p_pulse_stretch_timeout) THEN
                        o_fan_act_speed <= (OTHERS => '1'); -- Put -1 to indicate error
                        o_fan_error     <= '1';
                        sm_tacho_cnt    <= IDLE;
                     ELSE
                        IF (n_tacho_rising_edge = '1') THEN
                           v_tacho_period_cnt := (OTHERS => '0');
                           sm_tacho_cnt       <= WAIT_FOR_SECOND_TACHO_EDGE;
                        ELSE
                           v_tacho_period_cnt := v_tacho_period_cnt + 1;
                        END IF;
                     END IF;
                  -----------------------------------
                  WHEN WAIT_FOR_SECOND_TACHO_EDGE =>

                     IF (v_tacho_period_cnt >= p_pulse_stretch_timeout) THEN
                        o_fan_act_speed <= (OTHERS => '1'); -- Put -1 to indicate error
                        o_fan_error     <= '1';
                        sm_tacho_cnt    <= IDLE;
                     ELSE
                        IF (n_tacho_rising_edge = '1') THEN

                           -- Store result in array
                           v_measurements(v_measurement_index) := v_tacho_period_cnt;
                           IF (v_measurement_index = c_avg_length - 1) THEN
                              v_measurement_index := 0;
                           ELSE
                              v_measurement_index := v_measurement_index + 1;
                           END IF;
                           v_sum                               := (OTHERS => '0');

                           o_fan_error  <= '0';
                           sm_tacho_cnt <= CALC_AVG;
                        END IF;
                     END IF;

                     v_tacho_period_cnt := v_tacho_period_cnt + 1;
                  -----------------------------------
                  WHEN CALC_AVG =>

                     -- Accumulate measurements and divide to get avg
                     FOR i IN 0 TO c_avg_length - 1 LOOP
                        v_sum := v_sum + v_measurements(i);
                     END LOOP;
                     o_fan_act_speed <= std_logic_vector(v_sum / c_avg_length);

                     sm_tacho_cnt <= IDLE;

               END CASE;

            ELSE
               sm_tacho_cnt    <= IDLE;
               o_fan_act_speed <= (OTHERS => '0');
            END IF;

            v_do_speed_check_prev := r_do_speed_check;

         END IF;
      END IF;
   END PROCESS;

END ARCHITECTURE;

