LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;

PACKAGE topic_pkg IS

   -- ##########################################
   -- ########    TYPE DECLARATIONS    #########

   SUBTYPE t_byte IS std_logic_vector(7 DOWNTO 0);
   SUBTYPE t_word IS std_logic_vector(15 DOWNTO 0);
   TYPE t_byte_array IS ARRAY (natural RANGE <>) OF t_byte;
   TYPE t_word_array IS ARRAY (natural RANGE <>) OF t_word;

   TYPE t_slv8_array IS ARRAY (integer RANGE <>) OF std_logic_vector(7 DOWNTO 0);
   TYPE t_slv16_array IS ARRAY (integer RANGE <>) OF std_logic_vector(15 DOWNTO 0);
   TYPE t_slv32_array IS ARRAY (integer RANGE <>) OF std_logic_vector(31 DOWNTO 0);

   -- ########        END TYPES        #########
   -- ##########################################
   CONSTANT c_dna_width_7series    : integer := 57;
   CONSTANT c_dna_width_ultrascale : integer := 96;
   -- ##########################################
   -- ########    FUNCTION DECLARATIONS    #####

   FUNCTION sel(CON : BOOLEAN; IF_TRUE, IF_FALSE : INTEGER) RETURN INTEGER;
   FUNCTION sel_bit(CON : BOOLEAN; IF_TRUE, IF_FALSE : BIT) RETURN BIT;
   FUNCTION sel_string(CON : BOOLEAN; IF_TRUE, IF_FALSE : STRING) RETURN STRING;
   FUNCTION select_byte_from_slv(slv_in : std_logic_vector; byte_nr : natural) RETURN t_byte;
   FUNCTION set_byte_in_slv(slv_in : std_logic_vector; byte_nr : natural; byte_value : t_byte) RETURN std_logic_vector;
   FUNCTION log2ceil(L : POSITIVE) RETURN NATURAL;
   FUNCTION log2floor(L : POSITIVE) RETURN NATURAL;
   FUNCTION srlInteger(arg : integer; s : natural) RETURN integer;
   FUNCTION itoslv(arg : integer; width : integer) RETURN std_logic_vector;
   FUNCTION resize_slv(slv_in : std_logic_vector; size : natural) RETURN std_logic_vector;

   -- ########        END FUNCTIONS    #########
   -- ##########################################

   -- ###############################################
   -- ########    COMPONENT DECLARATIONS    #########

   COMPONENT xilinx_dna_axi_slave IS
      GENERIC(
         -- AXI bus related configuration parameters
         C_S_AXI_DATA_WIDTH  : integer := 32;
         C_S_AXI_ADDR_WIDTH  : integer := 32;
         -- Device architecture
         DEVICE_ARCHITECTURE : string  := "7SERIES"
      );
      PORT(
         -- ---------- AXI-4 Lite Interface -------
         S_AXI_ACLK    : IN  std_logic;
         S_AXI_ARESETN : IN  std_logic;
         S_AXI_AWADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_AWVALID : IN  std_logic;
         S_AXI_AWREADY : OUT std_logic;
         S_AXI_WDATA   : IN  std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_WSTRB   : IN  std_logic_vector((C_S_AXI_DATA_WIDTH / 8) - 1 DOWNTO 0);
         S_AXI_WVALID  : IN  std_logic;
         S_AXI_WREADY  : OUT std_logic;
         S_AXI_BRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_BVALID  : OUT std_logic;
         S_AXI_BREADY  : IN  std_logic;
         S_AXI_ARADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_ARVALID : IN  std_logic;
         S_AXI_ARREADY : OUT std_logic;
         S_AXI_RDATA   : OUT std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_RRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_RVALID  : OUT std_logic;
         S_AXI_RREADY  : IN  std_logic
      );
   END COMPONENT xilinx_dna_axi_slave;

   COMPONENT xilinx_dna_port IS
      GENERIC(
         -- Device architecture
         DEVICE_ARCHITECTURE : string := "7SERIES"
      );
      PORT(
         clk_in        : IN  std_logic;
         dna_valid_out : OUT std_logic := '0';
         dna_out       : OUT std_logic_vector(sel(DEVICE_ARCHITECTURE = "7SERIES", c_dna_width_7series, c_dna_width_ultrascale) - 1 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT single_clock_fifo IS
      GENERIC(
         FIFO_DEPTH         : natural RANGE 3 TO 12 := 10;
         FIFO_TYPE          : string                := "DISTRIBUTED"; -- May also be "BLOCK"
         FIFO_DATABUS_WIDTH : natural               := 32
      );
      PORT(
         fifo_clk      : IN  std_logic;
         fifo_rst      : IN  std_logic;
         fifo_wdata    : IN  std_logic_vector(FIFO_DATABUS_WIDTH - 1 DOWNTO 0);
         fifo_wr       : IN  std_logic;
         fifo_rdata    : OUT std_logic_vector(FIFO_DATABUS_WIDTH - 1 DOWNTO 0);
         fifo_rd       : IN  std_logic;
         fifo_empty    : OUT std_logic;
         fifo_full     : OUT std_logic;
         fifo_error    : OUT std_logic;
         fifo_vacant   : OUT std_logic_vector(12 DOWNTO 0);
         fifo_occupied : OUT std_logic_vector(12 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT dual_clock_fifo IS
      GENERIC(
         DATA_WIDTH : integer := 32;
         ADDR_WIDTH : integer := 4;
         FIFO_TYPE  : string  := "DISTRIBUTED"
      );
      PORT(
         -- Writing port
         wr_clk   : IN  std_logic;
         wr_reset : IN  std_logic;
         wr_data  : IN  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         wr_en    : IN  std_logic;
         full     : OUT std_logic;
         -- Reading port
         rd_clk   : IN  std_logic;
         rd_reset : IN  std_logic;
         rd_data  : OUT std_logic_vector(DATA_WIDTH - 1 DOWNTO 0) := (OTHERS => '0');
         rd_en    : IN  std_logic;
         empty    : OUT std_logic
      );
   END COMPONENT;

   COMPONENT dual_clock_fifo_lower_width IS
      GENERIC(
         DATA_WIDTH_IN  : integer := 128;
         DATA_WIDTH_OUT : integer := 32;
         ADDR_WIDTH     : integer := 4;
         FIFO_TYPE      : string  := "DISTRIBUTED"
      );
      PORT(
         -- Writing port
         wr_clk   : IN  std_logic;
         wr_reset : IN  std_logic;
         wr_data  : IN  std_logic_vector(DATA_WIDTH_IN - 1 DOWNTO 0);
         wr_en    : IN  std_logic;
         full     : OUT std_logic;
         -- Reading port
         rd_clk   : IN  std_logic;
         rd_reset : IN  std_logic;
         rd_data  : OUT std_logic_vector(DATA_WIDTH_OUT - 1 DOWNTO 0) := (OTHERS => '0');
         rd_en    : IN  std_logic;
         rd_last  : OUT std_logic;
         empty    : OUT std_logic
      );
   END COMPONENT;

   COMPONENT vector_cdc_hs IS
      GENERIC(
         VECTOR_WIDTH : integer := 10
      );
      PORT(
         -- input
         clk_in     : IN  std_logic;
         vector_in  : IN  std_logic_vector(VECTOR_WIDTH - 1 DOWNTO 0);
         captured   : OUT std_logic;
         -- output
         clk_out    : IN  std_logic;
         vector_out : OUT std_logic_vector(VECTOR_WIDTH - 1 DOWNTO 0);
         updated    : OUT std_logic
      );
   END COMPONENT;

   COMPONENT vector_cdc IS
      GENERIC(
         VECTOR_WIDTH  : integer   := 10;
         INITIAL_VALUE : std_logic := '0'
      );
      PORT(
         -- Asynchronous input
         vector_in  : IN  std_logic_vector(VECTOR_WIDTH - 1 DOWNTO 0);
         -- Synchronous output
         clk_out    : IN  std_logic;
         vector_out : OUT std_logic_vector(VECTOR_WIDTH - 1 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT std_logic_cdc IS
      GENERIC(
         INITIAL_VALUE : std_logic := '0'
      );
      PORT(
         -- Asynchronous input
         bit_in  : IN  std_logic;
         -- Synchronous output
         clk_out : IN  std_logic;
         bit_out : OUT std_logic
      );
   END COMPONENT;

   COMPONENT reset_cdc IS
      GENERIC(
         RESET_MIN_ACTIVE_WIDTH  : natural := 6;
         NUMBER_OF_CLOCK_DOMAINS : natural := 2
      );
      PORT(
         -- Synchronous reset in
         reset_clk_in : IN  std_logic;
         reset_in     : IN  std_logic;
         reset_out    : OUT std_logic                                              := '0';
         reset_done   : OUT std_logic                                              := '0';
         -- Synchronous resets out
         clks         : IN  std_logic_vector(NUMBER_OF_CLOCK_DOMAINS - 1 DOWNTO 0);
         resets_out   : OUT std_logic_vector(NUMBER_OF_CLOCK_DOMAINS - 1 DOWNTO 0) := (OTHERS => '0')
      );
   END COMPONENT;

   COMPONENT axil_slave_to_reg IS
      GENERIC(
         PNiRegs            : integer := 8;
         PNoRegs            : integer := 8;
         -- AXI bus related configuration parameters
         C_S_AXI_DATA_WIDTH : integer := 32;
         C_S_AXI_ADDR_WIDTH : integer := 32
      );
      PORT(
         -- ---------- AXI-4 Lite Interface -------
         S_AXI_ACLK    : IN  std_logic;
         S_AXI_ARESETN : IN  std_logic;
         S_AXI_AWADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_AWVALID : IN  std_logic;
         S_AXI_AWREADY : OUT std_logic;
         S_AXI_WDATA   : IN  std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_WSTRB   : IN  std_logic_vector(3 DOWNTO 0);
         S_AXI_WVALID  : IN  std_logic;
         S_AXI_WREADY  : OUT std_logic;
         S_AXI_BRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_BVALID  : OUT std_logic;
         S_AXI_BREADY  : IN  std_logic;
         S_AXI_ARADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_ARVALID : IN  std_logic;
         S_AXI_ARREADY : OUT std_logic;
         S_AXI_RDATA   : OUT std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_RRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_RVALID  : OUT std_logic;
         S_AXI_RREADY  : IN  std_logic;
         -- ---------- ADC Data input --------------
         iData         : IN  t_slv32_array(0 TO PNiRegs - 1);
         oData         : OUT t_slv32_array(0 TO PNoRegs - 1)
      );
   END COMPONENT;

   COMPONENT spi_master IS
      GENERIC(
         pSYSCLKFREQ      : integer := 100_000_000;
         pSPICLKFREQ      : integer := 1_000_000;
         pADDR_NBITS      : integer := 8;
         pDATA_NBITS      : integer := 16;
         pSAMPLECLKPERIOD : integer := 100_000;
         pSAMPLECLKHIGH   : integer := 50_000
      );
      PORT(
         -- Clock and Reset
         iSysRst      : IN  std_logic;
         iSysClk      : IN  std_logic;
         iSampleClkEn : IN  std_logic;
         -- Spi control and status interface
         iSpiData     : IN  std_logic_vector(pDATA_NBITS - 1 DOWNTO 0);
         iSpiAddr     : IN  std_logic_vector(pADDR_NBITS - 1 DOWNTO 0) := (OTHERS => '0');
         iSpiRead     : IN  std_logic;
         iSpiStart    : IN  std_logic;
         iSpiEn       : IN  std_logic;
         oSpiDone     : OUT std_logic;
         oSpiReadData : OUT std_logic_vector(pDATA_NBITS - 1 DOWNTO 0);
         -- Spi interface
         oSpiClk      : OUT std_logic;
         oSpiCs_n     : OUT std_logic;
         oSpiMosi     : OUT std_logic;
         iSpiMiso     : IN  std_logic;
         oSpiRead     : OUT std_logic;
         -- Sample clock
         oSampleClk   : OUT std_logic
      );
   END COMPONENT;

   COMPONENT spi_controller
      GENERIC(
         p_ADC_Width    : integer;
         p_ClkDiv       : integer;
         p_ADC_CHANNELS : integer
      );
      PORT(
         i_clk           : IN  std_logic;
         i_rst           : IN  std_logic;
         i_tx_start      : IN  std_logic;
         o_tx_end        : OUT std_logic;
         o_data_parallel : OUT std_logic_vector(p_ADC_Width - 1 DOWNTO 0);
         o_sclk          : OUT std_logic;
         o_ss            : OUT std_logic;
         o_mosi          : OUT std_logic;
         i_miso          : IN  std_logic
      );
   END COMPONENT spi_controller;

   COMPONENT spi_slave IS
      GENERIC(
         p_data_width : positive := 8
      );
      PORT(
         i_sys_clk         : IN  std_logic;
         -- spi interface
         i_sck             : IN  std_logic;
         i_cs              : IN  std_logic;
         i_mosi            : IN  std_logic;
         o_miso            : OUT std_logic                                   := '0';
         -- data interface
         o_miso_data_read  : OUT std_logic                                   := '0';
         i_miso_data       : IN  std_logic_vector(p_data_width - 1 DOWNTO 0) := (OTHERS => '0');
         o_mosi_data_valid : OUT std_logic                                   := '0';
         o_mosi_data       : OUT std_logic_vector(p_data_width - 1 DOWNTO 0) := (OTHERS => '0');
         o_spi_idle        : OUT std_logic                                   := '0'
      );
   END COMPONENT;

   COMPONENT axil_slave_to_ltc2508 IS
      GENERIC(
         pSYSCLKFREQ        : integer := 100_000_000;
         pSPICLKFREQ        : integer := 1_000_000;
         pADDR_NBITS        : integer := 0;
         pDATA_NBITS        : integer := 32;
         pSAMPLECLKPERIOD   : integer := 100_000; -- ns
         pSAMPLECLKHIGH     : integer := 50_000;
         C_S_AXI_DATA_WIDTH : integer := 32;
         C_S_AXI_ADDR_WIDTH : integer := 32); -- ns
      PORT(
         -- data from chip
         iDRL          : IN  std_logic;
         iBusy         : IN  std_logic;
         oIRQ          : OUT std_logic;
         -- spi signals
         oSpiClk       : OUT std_logic;
         oSpiCs_n      : OUT std_logic;
         oSpiMosi      : OUT std_logic;
         iSpiMiso      : IN  std_logic;
         -- data to chip
         oSampleClk    : OUT std_logic;
         -- ---------- AXI-4 Lite interface -------
         S_AXI_ACLK    : IN  std_logic;
         S_AXI_ARESETN : IN  std_logic;
         S_AXI_AWADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_AWVALID : IN  std_logic;
         S_AXI_AWREADY : OUT std_logic;
         S_AXI_WDATA   : IN  std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_WSTRB   : IN  std_logic_vector(3 DOWNTO 0);
         S_AXI_WVALID  : IN  std_logic;
         S_AXI_WREADY  : OUT std_logic;
         S_AXI_BRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_BVALID  : OUT std_logic;
         S_AXI_BREADY  : IN  std_logic;
         S_AXI_ARADDR  : IN  std_logic_vector(C_S_AXI_ADDR_WIDTH - 1 DOWNTO 0);
         S_AXI_ARVALID : IN  std_logic;
         S_AXI_ARREADY : OUT std_logic;
         S_AXI_RDATA   : OUT std_logic_vector(C_S_AXI_DATA_WIDTH - 1 DOWNTO 0);
         S_AXI_RRESP   : OUT std_logic_vector(1 DOWNTO 0);
         S_AXI_RVALID  : OUT std_logic;
         S_AXI_RREADY  : IN  std_logic);
   END COMPONENT;

   COMPONENT ltc2508_to_axis IS
      GENERIC(
         p_SYSCLKFREQ      : integer := 100_000_000;
         p_SPICLKFREQ      : integer := 1_000_000;
         p_ADDR_NBITS      : integer := 0;
         p_DATA_NBITS      : integer := 32;
         p_SAMPLECLKPERIOD : integer := 100_000;
         p_SAMPLECLKHIGH   : integer := 50_000
      );
      PORT(
         i_sys_clk     : IN  std_logic;
         i_sys_rst     : IN  std_logic;
         -- LTC2508 interface
         i_DRL         : IN  std_logic;
         i_busy        : IN  std_logic;
         o_irq         : OUT std_logic;
         o_sample_clk  : OUT std_logic;
         o_spi_clk     : OUT std_logic;
         o_spi_cs_n    : OUT std_logic;
         o_spi_mosi    : OUT std_logic;
         i_spi_miso    : IN  std_logic;
         -- Spi control
         i_control     : IN  std_logic_vector(1 DOWNTO 0);
         -- AXI-S output
         M_AXIS_TDATA  : OUT std_logic_vector(31 DOWNTO 0);
         M_AXIS_TVALID : OUT std_logic;
         M_AXIS_TREADY : IN  std_logic
      );
   END COMPONENT;

   COMPONENT serializer IS
      GENERIC(
         DATA_WIDTH          : natural RANGE 0 TO 8 := 4;
         DATA_RATE_INPUT     : string               := "SDR";
         -- Device architecture
         DEVICE_ARCHITECTURE : string               := "7SERIES"
      );
      port(
         i_clk            : in  std_logic;
         i_divclk         : in  std_logic;
         i_rst            : in  std_logic; -- asynchronous for ultrascale, synchronous to divclk for 7series
         parallel_data_in : in  std_logic_vector(DATA_WIDTH - 1 downto 0);
         serial_data_out  : out std_logic
      );
   END COMPONENT;

   COMPONENT deserializer IS
      GENERIC(
         DATA_WIDTH                          : natural RANGE 0 TO 8 := 4;
         DATA_RATE_OUTPUT                    : string               := "SDR";
         SYNC_AFTER_RESET                    : boolean              := true;
         SYNC_PATTERN                        : t_byte               := "11000000";
         CAPTURE_DIFF_INPUT_ON_OPPOSITE_EDGE : boolean              := false;
         -- Device architecture
         DEVICE_ARCHITECTURE                 : string               := "7SERIES";
         IODELAY_MODE                        : string               := "NONE"
      );
      port(
         i_clk             : in  std_logic;
         i_divclk          : in  std_logic;
         i_rst             : in  std_logic; -- asynchronous for ultrascale, synchronous to divclk for 7series
         serial_data_in    : in  std_logic;
         parallel_data_out : out std_logic_vector(DATA_WIDTH - 1 downto 0);
         in_sync           : out std_logic
      );
   END COMPONENT;

   COMPONENT input_delay IS
      GENERIC(
         DATA_WIDTH : natural := 1;
         DEFAULT_DELAY : integer := 0;
         DEVICE_ARCHITECTURE : string:= "7SERIES"
      );
      PORT(
        I  : in    std_logic_vector(DATA_WIDTH - 1 downto 0); --from input buffer 
        O  : out   std_logic_vector(DATA_WIDTH - 1 downto 0); --to FPGA or serializer
        C  : in    std_logic;                                 --Should be the same clock as clkdiv of serdes
        LD : in    std_logic;                                 --Update TAP delay, sync to C clock domain
        TAP_SELECT : in std_logic_vector(4 downto 0)          --number of tap selected,sync to C clock domain
     );
   END COMPONENT;

   COMPONENT diff_input_buffer IS
      PORT(
         input_p : IN  std_logic;
         input_n : IN  std_logic;
         output  : OUT std_logic
      );
   END COMPONENT;

   COMPONENT diff_output_buffer IS
      PORT(
         input    : IN  std_logic;
         output_p : OUT std_logic;
         output_n : OUT std_logic
      );
   END COMPONENT;

   COMPONENT diff_input_buffers IS
      GENERIC(
         DATA_WIDTH : natural := 1
      );
      PORT(
         input_p : IN  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         input_n : IN  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         output  : OUT std_logic_vector(DATA_WIDTH - 1 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT diff_output_buffers IS
      GENERIC(
         DATA_WIDTH : natural := 1
      );
      PORT(
         input    : IN  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         output_p : OUT std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         output_n : OUT std_logic_vector(DATA_WIDTH - 1 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT tristate_buffer IS
      GENERIC(
         DATA_WIDTH : natural := 1
      );
      PORT(
         I  : IN    std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         T  : IN    std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         O  : OUT   std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         IO : INOUT std_logic_vector(DATA_WIDTH - 1 DOWNTO 0)
      );
   END COMPONENT;

   COMPONENT pwm_fan_controller IS
      GENERIC(
         p_clkin_freq            : positive := 100_000_000; -- Frequency of i_clk in Hz
         p_pwm_freq              : positive := 30; -- Frequency of PWM signal in Hz
         p_pwm_precision         : positive := 255; -- Stepsize of PWM signal
         p_pulse_stretch_timeout : positive := 100_000_000 -- Timeout for pulse_stretching (if not tacho pulses are received)
      );
      PORT(
         i_clk           : IN  std_logic;
         i_rst           : IN  std_logic;
         i_fan_enable    : IN  std_logic;
         i_fan_req_speed : IN  std_logic_vector(7 DOWNTO 0); -- requested speed (in tacho pulses)
         o_fan_act_speed : OUT std_logic_vector(31 DOWNTO 0); -- actual speed (in tacho pulses)
         o_fan_pwm       : OUT std_logic;
         i_fan_tacho     : IN  std_logic;
         o_fan_error     : OUT std_logic
      );
   END COMPONENT;

   COMPONENT dsp_macc IS
      GENERIC(
         DATA_WIDTH : natural := 10;
         MULT_WIDTH : natural := 16;
         ACC_WIDTH  : natural := 16;
         OUT_WIDTH  : natural := 10
      );
      PORT(
         clk      : IN  std_logic;
         data_in  : IN  std_logic_vector(DATA_WIDTH - 1 DOWNTO 0);
         mult_in  : IN  std_logic_vector(MULT_WIDTH - 1 DOWNTO 0);
         acc_in   : IN  std_logic_vector(ACC_WIDTH - 1 DOWNTO 0);
         data_out : OUT std_logic_vector(OUT_WIDTH - 1 DOWNTO 0));
   END COMPONENT;

   COMPONENT crc_48 IS
      GENERIC(
         Nr_of_Bytes : integer := 128
      );
      PORT(
         iSysClk   : IN  std_logic;
         iSysRst   : IN  std_logic;
         crc_en    : IN  std_logic;
         data_in   : IN  std_logic_vector(15 DOWNTO 0);
         crc_out   : OUT std_logic_vector(47 DOWNTO 0);
         crc_valid : OUT std_logic
      );

   END COMPONENT;

   COMPONENT axi_i2c_controller IS
      GENERIC(
         ADRWIDTH         : integer := 15;
         p_axi_data_width : integer := 32
      );
      PORT(
         -- CLOCK AND RESET
         s_axi_aclk    : IN  std_logic;
         s_axi_aresetn : IN  std_logic;
         -- write address channel
         s_axi_awaddr  : IN  std_logic_vector(ADRWIDTH - 1 DOWNTO 0);
         s_axi_awvalid : IN  std_logic;
         s_axi_awready : OUT std_logic;
         -- write data channel
         s_axi_wdata   : IN  std_logic_vector(31 DOWNTO 0);
         s_axi_wstrb   : IN  std_logic_vector(3 DOWNTO 0);
         s_axi_wvalid  : IN  std_logic;
         s_axi_wready  : OUT std_logic;
         -- read address channel
         s_axi_araddr  : IN  std_logic_vector(ADRWIDTH - 1 DOWNTO 0);
         s_axi_arvalid : IN  std_logic;
         s_axi_arready : OUT std_logic;
         -- read data channel
         s_axi_rdata   : OUT std_logic_vector(31 DOWNTO 0);
         s_axi_rresp   : OUT std_logic_vector(1 DOWNTO 0);
         s_axi_rvalid  : OUT std_logic;
         s_axi_rready  : IN  std_logic;
         -- write response channel
         s_axi_bresp   : OUT std_logic_vector(1 DOWNTO 0);
         s_axi_bvalid  : OUT std_logic;
         s_axi_bready  : IN  std_logic;
         -- mclk
         mclk          : IN  std_logic;
         scl_padoen_oe : OUT std_logic;
         sda_padoen_oe : OUT std_logic;
         sda_pad_i     : IN  std_logic;
         scl_pad_i     : IN  std_logic;
         sda_pad_o     : OUT std_logic;
         scl_pad_o     : OUT std_logic
      );
   END COMPONENT;

   COMPONENT tristate_mux
      PORT(
         s_scl_i_0 : IN  std_logic;
         s_scl_i_1 : IN  std_logic;
         s_scl_t_0 : IN  std_logic;
         s_scl_t_1 : IN  std_logic;
         s_scl_o_0 : OUT std_logic;
         s_scl_o_1 : OUT std_logic;
         --
         s_sda_i_0 : IN  std_logic;
         s_sda_i_1 : IN  std_logic;
         s_sda_t_0 : IN  std_logic;
         s_sda_t_1 : IN  std_logic;
         s_sda_o_0 : OUT std_logic;
         s_sda_o_1 : OUT std_logic;
         --
         sel       : IN  std_logic;
         --
         m_scl_i   : IN  std_logic;
         m_sda_i   : IN  std_logic;
         m_scl_t   : OUT std_logic;
         m_sda_t   : OUT std_logic;
         m_scl_o   : OUT std_logic;
         m_sda_o   : OUT std_logic
      );
   END COMPONENT;

   COMPONENT i2c_master_top IS
      GENERIC(
         ARST_LVL : std_logic := '0'    -- asynchronous reset level
      );
      PORT(
         -- wishbone signals
         wb_clk_i     : IN  std_logic;  -- master clock input
         wb_rst_i     : IN  std_logic := '0'; -- synchronous active high reset
         wb_adr_i     : IN  std_logic_vector(2 DOWNTO 0); -- lower address bits
         wb_dat_i     : IN  std_logic_vector(31 DOWNTO 0); -- Databus input
         wb_dat_o     : OUT std_logic_vector(31 DOWNTO 0); -- Databus output
         wb_we_i      : IN  std_logic;  -- Write enable input
         wb_stb_i     : IN  std_logic;  -- Strobe signals / core select signal
         wb_cyc_i     : IN  std_logic;  -- Valid bus cycle input
         wb_ack_o     : OUT std_logic;  -- Bus cycle acknowledge output
         wb_inta_o    : OUT std_logic;  -- interrupt request output signal
         mclk         : IN  std_logic;
         -- i2c lines
         scl_pad_i    : IN  std_logic;  -- i2c clock line input
         scl_pad_o    : OUT std_logic;  -- i2c clock line output
         scl_padoen_o : OUT std_logic;  -- i2c clock line output enable, active low
         sda_pad_i    : IN  std_logic;  -- i2c data line input
         sda_pad_o    : OUT std_logic;  -- i2c data line output
         sda_padoen_o : OUT std_logic   -- i2c data line output enable, active low
      );
   END COMPONENT;

   COMPONENT i2c_master_byte_ctrl IS
      PORT(
         clk      : IN  std_logic;
         rst      : IN  std_logic;
         ena      : IN  std_logic;
         mclk     : IN  std_logic;
         clk_cnt  : IN  std_logic_vector(15 DOWNTO 0); -- 4x SCL

         -- input signals
         start    : IN  std_logic;
         stop     : IN  std_logic;
         read     : IN  std_logic;
         write    : IN  std_logic;
         ack_in   : IN  std_logic;
         din      : IN  std_logic_vector(7 DOWNTO 0);
         -- output signals
         cmd_ack  : OUT std_logic;
         ack_out  : OUT std_logic;
         i2c_busy : OUT std_logic;
         i2c_al   : OUT std_logic;
         dout     : OUT std_logic_vector(7 DOWNTO 0);
         -- i2c lines
         scl_i    : IN  std_logic;      -- i2c clock line input
         scl_o    : OUT std_logic;      -- i2c clock line output
         scl_oen  : OUT std_logic;      -- i2c clock line output enable, active low
         sda_i    : IN  std_logic;      -- i2c data  line input
         sda_o    : OUT std_logic;      -- i2c data  line output
         sda_oen  : OUT std_logic       -- i2c data  line output enable, active low
      );
   END COMPONENT;


   -- ########   END COMPONENT DECLARATIONS   #######
   -- ###############################################

END PACKAGE;

PACKAGE BODY topic_pkg IS
   FUNCTION sel(CON : BOOLEAN; IF_TRUE, IF_FALSE : INTEGER) RETURN INTEGER IS
   BEGIN
      IF CON = TRUE THEN
         RETURN IF_TRUE;
      ELSE
         RETURN IF_FALSE;
      END IF;
   END sel;
   
   FUNCTION sel_bit(CON : BOOLEAN; IF_TRUE, IF_FALSE : BIT) RETURN BIT IS
   BEGIN
      IF CON = TRUE THEN
         RETURN IF_TRUE;
      ELSE
         RETURN IF_FALSE;
      END IF;
   END sel_bit;
   
   FUNCTION sel_string(CON : BOOLEAN; IF_TRUE, IF_FALSE : STRING) RETURN STRING IS
   BEGIN
      IF CON = TRUE THEN
         RETURN IF_TRUE;
      ELSE
         RETURN IF_FALSE;
      END IF;
   END sel_string;     

   FUNCTION select_byte_from_slv(slv_in : std_logic_vector; byte_nr : natural) RETURN t_byte IS
   BEGIN
      IF (byte_nr < slv_in'length / 8) THEN
         RETURN t_byte(slv_in(((byte_nr + 1) * 8) - 1 DOWNTO (byte_nr * 8)));
      END IF;
   END FUNCTION;

   FUNCTION set_byte_in_slv(slv_in : std_logic_vector; byte_nr : natural; byte_value : t_byte) RETURN std_logic_vector IS
      VARIABLE slv_tmp : std_logic_vector(slv_in'RANGE);
   BEGIN
      slv_tmp                                               := slv_in;
      slv_tmp(((byte_nr + 1) * 8) - 1 DOWNTO (byte_nr * 8)) := byte_value;
      RETURN slv_tmp;
   END FUNCTION;

   FUNCTION log2ceil(L : POSITIVE) RETURN NATURAL IS
      VARIABLE v_i        : natural;
      VARIABLE v_bitCount : natural;
   BEGIN
      v_i        := L - 1;
      v_bitCount := 0;
      WHILE (v_i > 0) LOOP
         v_bitCount := v_bitCount + 1;
         v_i        := srlInteger(v_i, 1);
      END LOOP;
      RETURN v_bitCount;
   END log2ceil;

   FUNCTION log2floor(L : POSITIVE) RETURN NATURAL IS
      VARIABLE v_i        : natural;
      VARIABLE v_bitCount : natural;
   BEGIN
      v_i        := L;
      v_bitCount := 0;
      WHILE (v_i > 1) LOOP
         v_bitCount := v_bitCount + 1;
         v_i        := srlInteger(v_i, 1);
      END LOOP;
      RETURN v_bitCount;
   END log2floor;

   FUNCTION srlInteger(arg : integer; s : natural) RETURN integer IS
   BEGIN
      RETURN to_integer(SHIFT_RIGHT(to_UNSIGNED(arg, 32), s));
   END srlInteger;

   -- Function to convert easily an integer value to std_logic_vector
   -- arg = the integer value
   -- width = the vector width to convert to
   -- return: std_logic_vector(width-1 downto 0);
   FUNCTION itoslv(arg : integer; width : integer) RETURN std_logic_vector IS
   BEGIN
      RETURN std_logic_vector(to_unsigned(arg, width));
   END itoslv;

   FUNCTION resize_slv(slv_in : std_logic_vector; size : natural) RETURN std_logic_vector IS
   BEGIN
      RETURN std_logic_vector(resize(unsigned(slv_in), size));
   END resize_slv;

END PACKAGE BODY;
